/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.interceptor;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.petals.component.framework.interceptor.description.InterceptorConfiguration;

/**
 * One message interceptor manager per Consume / Provide extension.
 *
 * @author Christophe HAMERLING (chamerling) - eBM WebSourcing
 * @since PCF 2.0
 *
 */
public class MessageInterceptorManager {

    /**
     * The input interceptors to be used. The Map is ordered, the iterator used
     * to iterate over the entries will return the {@link Interceptor} in the
     * user defined order.
     */
    protected Map<String, Interceptor> inputOrderedInterceptors;
    
    /**
     * The input accept interceptors to be used. The Map is ordered, the
     * iterator used to iterate over the entries will return the
     * {@link Interceptor} in the user defined order.
     */
    protected Map<String, Interceptor> inputResponseOrderedInterceptors;

    /**
     * The output interceptors to be used. The Map is ordered, the iterator used
     * to iterate over the entries will return the {@link Interceptor} in the
     * user defined order.
     */
    protected Map<String, Interceptor> outputOrderedInterceptors;
    
    /**
     * The output response interceptors to be used. The Map is ordered, the
     * iterator used to iterate over the entries will return the
     * {@link Interceptor} in the user defined order.
     */
    protected Map<String, Interceptor> outputResponseOrderedInterceptors;

    /**
     * Additional configuration at SU level for input interceptors
     */
    protected Map<String, InterceptorConfiguration> inputAdditionalConfiguration;
    
    /**
     * Additional configuration at SU level for input response interceptors
     */
    protected Map<String, InterceptorConfiguration> inputResponseAdditionalConfiguration;

    /**
     * Additional configuration at SU level for output interceptors
     */
    protected Map<String, InterceptorConfiguration> outputAdditionalConfiguration;
    
    /**
     * Additional configuration at SU level for output response interceptors
     */
    protected Map<String, InterceptorConfiguration> outputResponseAdditionalConfiguration;

    /**
     * Logger
     */
    protected Logger logger;

    /**
     * Creates a new instance of {@link MessageInterceptorManager}
     *
     * @param logger
     */
    public MessageInterceptorManager(final Logger logger) {
        this.logger = logger;
        this.inputOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.inputResponseOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.outputOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.outputResponseOrderedInterceptors = new LinkedHashMap<String, Interceptor>();

        this.inputAdditionalConfiguration = new HashMap<String, InterceptorConfiguration>();
        this.inputResponseAdditionalConfiguration = new HashMap<String, InterceptorConfiguration>();
        this.outputAdditionalConfiguration = new HashMap<String, InterceptorConfiguration>();
        this.outputResponseAdditionalConfiguration = new HashMap<String, InterceptorConfiguration>();
    }

    /**
     * Initialization
     *
     * @param componentInterceptors
     * @param localConfiguration
     * @param inOrder
     * @param outOrder
     */
    public void init(final Map<String, Interceptor> componentInterceptors,
            final Map<String, InterceptorConfiguration> inConfiguration,
            final Map<String, InterceptorConfiguration> outConfiguration,
            final Map<String, InterceptorConfiguration> inResponseConfiguration,
            final Map<String, InterceptorConfiguration> outResponseConfiguration) {
        logger.fine("Initializing interceptor manager");

        logger.fine("Initializing interceptor manager");

        // accept
        this.initInterceptor(componentInterceptors, inConfiguration, inputOrderedInterceptors,
                inputAdditionalConfiguration);
        this.initInterceptor(componentInterceptors, inResponseConfiguration,
                inputResponseOrderedInterceptors, inputResponseAdditionalConfiguration);

        // send
        this.initInterceptor(componentInterceptors, outConfiguration, outputOrderedInterceptors,
                outputAdditionalConfiguration);
        this.initInterceptor(componentInterceptors, outResponseConfiguration,
                outputResponseOrderedInterceptors, outputResponseAdditionalConfiguration);
    }

    /**
     * Initialize the interceptor manager. Put all the interceptors in the good
     * order.
     *
     * @param componentInterceptor
     *            the JBI component level interceptors. Must be non null.
     * @param localInterceptor
     *            the interceptors to be setted
     * @param orderedInterceptorsConfiguration
     *            the interceptors configuration
     */
    protected void initInterceptor(
            final Map<String, Interceptor> componentInterceptors,
            final Map<String, InterceptorConfiguration> originalInterceptorsConfiguration,
            Map<String, Interceptor> orderedInterceptors,
            Map<String, InterceptorConfiguration> additionalInterceptorsConfiguration) {

        logger.fine("Initializing interceptors...");

        // set additional configuration
        additionalInterceptorsConfiguration.putAll(originalInterceptorsConfiguration);

        // add the component active interceptors in the component defined
        // order. The component interceptors have higher priority
        Set<String> keys = componentInterceptors.keySet();
        for (String itcName : keys) {
            Interceptor interceptor = componentInterceptors.get(itcName);
            if (interceptor.isActive()) {
                orderedInterceptors.put(interceptor.getName(), interceptor);
            }
        }

        if (originalInterceptorsConfiguration != null) {

            // add the SU interceptors (not the already added ones)
            keys = originalInterceptorsConfiguration.keySet();
            for (String itcName : keys) {

                // get the component interceptor if it exists
                Interceptor interceptor = componentInterceptors.get(itcName);
                if (interceptor != null && orderedInterceptors.get(itcName) == null) {
                    orderedInterceptors.put(interceptor.getName(), interceptor);
                } else {
                    logger
                            .warning("Interceptor "
                                    + itcName
                                    + " not found in component interceptors. It will not be used in this manager");
                }
            }
        }
    }

    /**
     * @return the inputOrderedInterceptors
     */
    public Map<String, Interceptor> getInputOrderedInterceptors() {
        return this.inputOrderedInterceptors;
    }

    /**
     * @return the outputOrderedInterceptors
     */
    public Map<String, Interceptor> getOutputOrderedInterceptors() {
        return this.outputOrderedInterceptors;
    }

    /**
     * @return the inputAdditionalConfiguration
     */
    public Map<String, InterceptorConfiguration> getInputAdditionalConfiguration() {
        return this.inputAdditionalConfiguration;
    }

    /**
     * @return the outputAdditionalConfiguration
     */
    public Map<String, InterceptorConfiguration> getOutputAdditionalConfiguration() {
        return this.outputAdditionalConfiguration;
    }
    
    /**
     * @return the inputResponseOrderedInterceptors
     */
    public Map<String, Interceptor> getInputResponseOrderedInterceptors() {
        return this.inputResponseOrderedInterceptors;
    }

    /**
     * @return the outputResponseOrderedInterceptors
     */
    public Map<String, Interceptor> getOutputResponseOrderedInterceptors() {
        return this.outputResponseOrderedInterceptors;
    }

    /**
     * @return the inputResponseAdditionalConfiguration
     */
    public Map<String, InterceptorConfiguration> getInputResponseAdditionalConfiguration() {
        return this.inputResponseAdditionalConfiguration;
    }

    /**
     * @return the outputResponseAdditionalConfiguration
     */
    public Map<String, InterceptorConfiguration> getOutputResponseAdditionalConfiguration() {
        return this.outputResponseAdditionalConfiguration;
    }
}
