/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.util;

import java.net.URI;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.framework.MEPConstants;
import org.objectweb.petals.component.framework.PEtALSCFException;

/**
 * This is a Message Exchange Wrapper. The "in" and "out" normalized message are
 * cached into this class, authorizing the reset of these message. Can be use by
 * the Component Framework interceptors.
 *
 * @author ofabre
 *
 */
public class ExchangeImpl implements Exchange {

    private MessageExchange exchange;

    private static final String IN_MESSAGE_NAME = "in";

    private static final String OUT_MESSAGE_NAME = "out";
    
    private static final QName DEFAULT_FAULT_QNAME = new QName("0");

    private NormalizedMessage inNormalizedExchange;

    private NormalizedMessage outNormalizedExchange;

    private boolean checkRole;

    /**
     *
     * Creates a new instance of {@link ExchangeImpl}
     *
     * @param exchange
     */
    public ExchangeImpl(MessageExchange exchange) {
        super();
        this.exchange = exchange;
        this.inNormalizedExchange = this.exchange.getMessage(IN_MESSAGE_NAME);
        this.outNormalizedExchange = this.exchange.getMessage(OUT_MESSAGE_NAME);
        this.checkRole = true;
    }

    /**
     * Creates a message
     *
     * @return
     * @throws MessagingException
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return this.exchange.createMessage();
    }

    /**
     * Return the Wrapper Message Exchange. Set the "in" or the "out" normalized
     * messages if they are not already set.
     *
     * @return
     */
    public MessageExchange getMessageExchange() {
        try {
            if ((this.inNormalizedExchange != null)
                    && (this.exchange.getMessage(IN_MESSAGE_NAME) == null)) {
                this.exchange.setMessage(this.inNormalizedExchange,
                        IN_MESSAGE_NAME);
            }

            if ((this.outNormalizedExchange != null)
                    && (this.exchange.getMessage(OUT_MESSAGE_NAME) == null)) {
                this.exchange.setMessage(this.outNormalizedExchange,
                        OUT_MESSAGE_NAME);
            }
        } catch (MessagingException e) {
            // do nothing
        }

        return this.exchange;
    }

    /*
     * (non-Javadoc)
     * @see org.objectweb.petals.component.framework.util.Exchange#createFault()
     */
    public Fault createFault() throws MessagingException {
        return this.exchange.createFault();
    }

    /*
     * (non-Javadoc)
     * @see org.objectweb.petals.component.framework.util.Exchange#getEndpoint()
     */
    public ServiceEndpoint getEndpoint() {
        return this.exchange.getEndpoint();
    }

    /*
     * (non-Javadoc)
     * @see org.objectweb.petals.component.framework.util.Exchange#getError()
     */
    public Exception getError() {
        return this.exchange.getError();
    }

    public String getExchangeId() {
        return this.exchange.getExchangeId();
    }

    public Fault getFault() {
        return this.exchange.getFault();
    }

    public QName getInterfaceName() {
        return this.exchange.getInterfaceName();
    }

    public QName getOperation() {
        return this.exchange.getOperation();
    }

    public URI getPattern() {
        return this.exchange.getPattern();
    }

    public Object getProperty(String name) {
        return this.exchange.getProperty(name);
    }

    public Set getPropertyNames() {
        return this.exchange.getPropertyNames();
    }

    public MessageExchange.Role getRole() {
        return this.exchange.getRole();
    }

    public QName getService() {
        return this.exchange.getService();
    }

    public ExchangeStatus getStatus() {
        return this.exchange.getStatus();
    }

    public boolean isTransacted() {
        return this.exchange.isTransacted();
    }

    public void setEndpoint(ServiceEndpoint endpoint) {
        this.exchange.setEndpoint(endpoint);
    }

    /**
     * Put the given exception as Error part of the message exchange. Allowed
     * exception types are : {@link java.*}, {@link javax.jbi.JBIException},
     * org.w3c.* and org.xml.*.
     *
     */
    public void setError(Exception error) {
        String packageName = error.getClass().getPackage().getName();
        if (packageName.startsWith("java") || packageName.startsWith("org.w3c")
                || packageName.startsWith("org.xml")) {
            this.exchange.setError(error);
        } else {
            throw new IllegalArgumentException("Exception of type ("
                    + packageName + ") are not supported by the JBI container");
        }
    }

    public void setFault(Fault fault) throws MessagingException {
        this.exchange.setFault(fault);
    }

    public void setInterfaceName(QName interfaceName) {
        this.exchange.setInterfaceName(interfaceName);
    }

    public void setOperation(QName name) {
        this.exchange.setOperation(name);
    }

    public void setProperty(String name, Object obj) {
        this.exchange.setProperty(name, obj);
    }

    public void setService(QName service) {
        this.exchange.setService(service);
    }

    public void setStatus(ExchangeStatus status) throws MessagingException {
        this.exchange.setStatus(status);
    }

    public void setOutMessageContent(String outContent)
            throws MessagingException {
        try {
            NormalizedMessage outMess = this.getOutMessage();
            if (outMess == null) {
                outMess = this.exchange.createMessage();
            }
            outMess.setContent(SourceHelper.createSourceFromString(outContent));
            this.setOutMessage(outMess);
        } catch (PEtALSCFException e) {
            throw new MessagingException("Error creating response message");
        }
    }

    public void setOutMessageContent(Source outContent)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.setContent(outContent);
        this.setOutMessage(outMess);
    }

    public void setInMessageContent(String inContent) throws MessagingException {
        try {
            NormalizedMessage inMess = this.getInMessage();
            if (inMess == null) {
                inMess = this.exchange.createMessage();
            }
            inMess.setContent(SourceHelper.createSourceFromString(inContent));
            this.setInMessage(inMess);
        } catch (PEtALSCFException e) {
            throw new MessagingException("Error creating request message");
        }
    }

    public void setInMessageContent(Source inContent) throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.setContent(inContent);
        this.setInMessage(inMess);
    }

    public void setOutMessageAttachements(Map<String, DataHandler> attachements)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        for (String attachementId : attachements.keySet()) {
            outMess.addAttachment(attachementId, attachements
                    .get(attachementId));
        }
        this.setOutMessage(outMess);
    }

    public String getInMessageContentAsString() throws MessagingException {
        NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message must not be null");
        }
        Source content = inMessage.getContent();
        String stringContent = null;
        try {
            stringContent = SourceHelper.createString(content);
        } catch (PEtALSCFException e) {
            throw new MessagingException("Error creating string from source", e);
        }
        return stringContent;
    }

    public Source getInMessageContentAsSource() throws MessagingException {
        NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message must not be null");
        }
        return inMessage.getContent();
    }

    public String getOutMessageContentAsString() throws MessagingException {
        NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message must not be null");
        }
        Source content = outMessage.getContent();
        String stringContent = null;
        try {
            stringContent = SourceHelper.createString(content);
        } catch (PEtALSCFException e) {
            throw new MessagingException("Error creating string from source", e);
        }
        return stringContent;
    }

    public Source getOutMessageContentAsSource() throws MessagingException {
        NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message must not be null");
        }
        return outMessage.getContent();
    }

    public Set<DataHandler> getInMessageAttachments() throws MessagingException {
        Set<DataHandler> attachments = new HashSet<DataHandler>();

        NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message must not be null");
        }

        Set names = inMessage.getAttachmentNames();
        for (Object attachId : names) {
            DataHandler dh = inMessage.getAttachment((String) attachId);
            if (dh != null)
                attachments.add(dh);
        }
        return attachments;
    }

    public URI getExchangePattern() throws MessagingException {
        URI pattern = this.exchange.getPattern();
        if (pattern == null) {
            throw new MessagingException(
                    "You must specify a message exchange pattern in the message exchange");
        }
        return pattern;
    }

    public String getOperationName() throws MessagingException {
        QName operation = this.exchange.getOperation();
        if (operation == null) {
            throw new MessagingException(
                    "You must specify an operation the message exchange");
        }
        return operation.getLocalPart();
    }

    public String getEndpointName() throws MessagingException {
        ServiceEndpoint ep = this.exchange.getEndpoint();
        if (ep == null) {
            throw new MessagingException(
                    "You must specify an endpoint in the message exchange");
        }
        return ep.getEndpointName();
    }

    public NormalizedMessage getInMessage() {
        if (this.inNormalizedExchange == null) {
            if (this.exchange.getMessage(IN_MESSAGE_NAME) != null) {
                return this.exchange.getMessage(IN_MESSAGE_NAME);
            }
            try {
                this.inNormalizedExchange = this.exchange.createMessage();
            } catch (MessagingException e) {
                return null;
            }
        }
        return this.inNormalizedExchange;
    }
    
    public boolean isInMessage() {
        return this.exchange.getMessage(IN_MESSAGE_NAME) != null || this.inNormalizedExchange != null;
    }

    public void setInMessage(final NormalizedMessage msg)
            throws MessagingException {
        this.checkPatternMatching(IN_MESSAGE_NAME);
        this.checkRoleMatching(IN_MESSAGE_NAME);
        this.inNormalizedExchange = msg;
    }

    public NormalizedMessage getOutMessage() {
        if (this.outNormalizedExchange == null) {
            if (this.exchange.getMessage(OUT_MESSAGE_NAME) != null) {
                return this.exchange.getMessage(OUT_MESSAGE_NAME);
            } else {
                try {
                    this.outNormalizedExchange = this.exchange.createMessage();
                } catch (MessagingException e) {
                    return null;
                }
            }
        }
        return this.outNormalizedExchange;
    }
    
    public boolean isOutMessage() {
        return this.exchange.getMessage(OUT_MESSAGE_NAME) != null || this.outNormalizedExchange != null;
    }

    public void setOutMessage(final NormalizedMessage msg)
            throws MessagingException {
        this.checkPatternMatching(OUT_MESSAGE_NAME);
        this.checkRoleMatching(OUT_MESSAGE_NAME);
        this.outNormalizedExchange = msg;
    }

    /* ### STATUS ### */

    public void setDoneStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.DONE);
    }

    public boolean isDoneStatus() {
        return (ExchangeStatus.DONE.equals(this.getStatus()));
    }

    public void setErrorStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ERROR);
    }

    public boolean isErrorStatus() {
        return (ExchangeStatus.ERROR.equals(this.getStatus()));
    }

    public void setActiveStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ACTIVE);
    }

    public boolean isActiveStatus() {
        return (ExchangeStatus.ACTIVE.equals(this.getStatus()));
    }

    /* ### ROLE ### */

    public boolean isProviderRole() {
        return (MessageExchange.Role.PROVIDER.equals(this.getRole()));
    }

    public boolean isConsumerRole() {
        return (MessageExchange.Role.CONSUMER.equals(this.getRole()));
    }

    /* ### PATTERN ### */

    public boolean isInOnlyPattern() {
        return (MEPConstants.IN_ONLY_PATTERN.value().equals(this.exchange
                .getPattern()));
    }

    public boolean isRobustInOnlyPattern() {
        return (MEPConstants.ROBUST_IN_ONLY_PATTERN.value()
                .equals(this.exchange.getPattern()));
    }

    public boolean isInOutPattern() {
        return (MEPConstants.IN_OUT_PATTERN.value().equals(this.exchange
                .getPattern()));
    }

    public boolean isInOptionalOutPattern() {
        return (MEPConstants.IN_OPTIONAL_OUT_PATTERN.value()
                .equals(this.exchange.getPattern()));
    }

    public void setFault(Exception e) throws MessagingException {
        Fault f = this.exchange.createFault();
        final String faultString = SourceHelper.createSoapFault(e,
                ExchangeImpl.DEFAULT_FAULT_QNAME);
        Source content;
        try {
            content = SourceHelper.createSourceFromString(faultString);
        } catch (PEtALSCFException e1) {
            throw new MessagingException(
                    "Can't create Fault content from Exception", e1);
        }
        f.setContent(content);
        this.exchange.setFault(f);
    }

    public void setCheckRole(boolean checkRole) {
        this.checkRole = checkRole;
    }

    /**
     * Check that the exchange MEP allows the specified reference. This only
     * works for the JBI predefined MEPs. It check that reference is IN or OUT.
     * Other reference or MEP will cause an exception
     *
     * @param name
     * @throws MessagingException
     *             the MEP does not allow the reference
     */
    protected void checkPatternMatching(String name) throws MessagingException {
        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (OUT_MESSAGE_NAME.equalsIgnoreCase(name)
                && ((MEPConstants.valueOf(this.exchange.getPattern()).equals(
                        MEPConstants.IN_ONLY_PATTERN) || MEPConstants.valueOf(
                        this.exchange.getPattern()).equals(
                        MEPConstants.ROBUST_IN_ONLY_PATTERN)))) {
            throw new MessagingException(
                    "the MessageExchange state does not allow this operation.");
        }
    }

    /**
     * Check that the Role allows the specified reference. This only works for
     * the JBI predefined MEPs. It check that reference is IN or OUT and the
     * Role is CONSUMER or PROVIDER. Other reference will cause an exception
     *
     * @param name
     * @throws MessagingException
     *             the Role does not allow the reference
     */
    protected void checkRoleMatching(String name) throws MessagingException {
        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        // The check on the role can be unactivated for PCF intercepting purpose
        if (!this.checkRole) {
            return;
        }

        if (MessageExchange.Role.CONSUMER.equals(this.exchange.getRole())) {
            if (IN_MESSAGE_NAME.equalsIgnoreCase(name)) {
                return;
            }
        } else if (MessageExchange.Role.PROVIDER
                .equals(this.exchange.getRole())
                && OUT_MESSAGE_NAME.equalsIgnoreCase(name)) {
            return;
        }
        throw new MessagingException("The Role does not allow this operation.");
    }


}
