/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SourceHelper.java 94 2006-03-26 17:11:05Z alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.util;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.util.Properties;

import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.petals.component.framework.PEtALSCFException;

/**
 * Sample class used to manage the mapping between
 * <code>javax.xml.transform.Source</code> from String. the mapping is a
 * quicly "string" -> "<content>string</content>" conversion. Initial
 * developer(s): Adrien LOUIS
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public final class SourceHelper {

    private static final String START_TAG = "<content>";

    private static final String END_TAG = "</content>";

    private static final int SOAP_FAULT_BUFFER_SIZE = 2048;

    private static Transformer transformer;

    /**
     * Creates a new instance of {@link SourceHelper}
     * 
     */
    private SourceHelper() {
        super();
    }

    /**
     * Create a Source from String adding start and end tags.
     * 
     * @param msg
     * @return
     * @throws PEtALSCFException
     */
    public static Source createContentSource(final String msg) throws PEtALSCFException {
        return createSource(START_TAG + msg + END_TAG);
    }

    /**
     * Creates an XML fault with the stacktrace, message and the specified code
     * 
     * @param e
     *            exception, used to provide faultstring and detail
     * @param code
     *            code of the fault
     * @return A string containing the xml code that corresponds to the Soap
     *         Fault
     * @deprecated use method createSoapFault(final Throwable e, final QName
     *             qname) instead
     */
    public static String createSoapFault(final Throwable e, final String code) {
        return SourceHelper.createSoapFault(e, new QName(code));
    }

    /**
     * Creates an XML fault with the stacktrace, message and the specified code
     * 
     * @param e
     *            exception, used to provide faultstring and detail
     * @param qname
     *            qname of the fault
     * @return A string containing the xml code that corresponds to the Soap
     *         Fault
     */
    public static String createSoapFault(final Throwable e, final QName qname) {
        StringBuffer s = new StringBuffer(SourceHelper.SOAP_FAULT_BUFFER_SIZE);
        s.append("<s:Fault xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">");
        s.append("<faultcode>");
        if (qname != null) {
            s.append(qname);
        }
        s.append("</faultcode><faultstring>");
        s.append(e.getMessage());
        s.append("</faultstring><detail><![CDATA[");
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        s.append(sw);
        s.append("]]></detail></s:Fault>");
        return s.toString();
    }

    /**
     * Create a Source from the String. The string must be in UTF-8 format,
     * otherwise, the Source is not created.
     * 
     * @param msg
     *            an UTF-8 String
     * @return the Source built from the String, null if not an UTF-8
     * @deprecated use method createSourceFromString()
     */
    @Deprecated
    public static Source createSource(final String msg) throws PEtALSCFException {
        StreamSource source = new StreamSource();
        try {
            byte[] msgByte = msg.getBytes("UTF-8");
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (IOException e) {
            throw new PEtALSCFException(e);
        }
        return source;
    }

    /**
     * Create a Source from the String. The String must be a valid XML String.
     * If an encoding is specified in the XML String header it is used to create
     * the XML bytes stream, otherwise the default UTF-8 encoding is used.
     * 
     * @param msg
     *            a valid xml String
     * @return the Source built from the XML String
     * @throws UnsupportedEncodingException
     *             If the named charset is not supported
     * @throws PEtALSCFException
     *             if an error occurs during Source creation
     */
    public static Source createSourceFromString(final String msg) throws PEtALSCFException {
        String charsetEncoding = XMLUtil.extractXmlEncoding(msg);
        return createSourceFromString(msg, charsetEncoding);
    }

    /**
     * Create a Source from the String with the given charset encoding. The
     * String must be a valid XML String.
     * 
     * @param xmlMessage
     *            a valid xml String message
     * @param charsetEncoding
     *            a supported encoding
     * @return the Source built from the XML String with the specified encoding
     * @throws UnsupportedEncodingException
     *             If the named charset is not supported
     * @throws PEtALSCFException
     *             if an error occurs during Source creation
     */
    private static Source createSourceFromString(final String xmlMessage, String charsetEncoding)
            throws PEtALSCFException {
        StreamSource source = new StreamSource();
        try {
            byte[] msgByte = xmlMessage.getBytes(charsetEncoding);
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (IOException e) {
            throw new PEtALSCFException(e);
        }
        return source;
    }

    /**
     * Create a String from a Source. The String will be in UTF-8.
     * 
     * @param s
     * @return
     * @throws PEtALSCFException
     */
    public static String createString(final Source s) throws PEtALSCFException {

        String result = null;
        try {
            if (s instanceof StreamSource && ((StreamSource) s).getInputStream().markSupported()) {
                StreamSource ss = (StreamSource) s;
                ss.getInputStream().reset();
                byte[] buf = new byte[ss.getInputStream().available()];
                BufferedInputStream bis = new BufferedInputStream(ss.getInputStream());
                bis.read(buf);
                result = new String(buf);
                ss.getInputStream().reset();
            } else {
                StringWriter buffer = new StringWriter();
                Result sresult = new StreamResult(buffer);
                Transformer transform = getTransformer();
                transform.transform(s, sresult);
                result = buffer.toString();
            }
        } catch (IOException e) {
            throw new PEtALSCFException(e);
        } catch (TransformerConfigurationException e) {
            throw new PEtALSCFException(e);
        } catch (TransformerException e) {
            throw new PEtALSCFException(e);
        }
        return result;
    }

    /**
     * 
     * @return
     * @throws TransformerConfigurationException
     */
    protected synchronized static Transformer getTransformer()
            throws TransformerConfigurationException {
        if (transformer == null) {
            transformer = TransformerFactory.newInstance().newTransformer();
            Properties props = new Properties();
            props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperties(props);
        }
        return transformer;
    }

    /**
     * Create a String from a Source.
     * 
     * @param s
     * @return the String in UTF-8 representation without the start and end tags
     * @throws PEtALSCFException
     */
    public static String createStringContent(final Source s) throws PEtALSCFException {
        String content = createString(s);
        int index = content.indexOf(START_TAG);
        String result = content.substring(index + START_TAG.length());
        return result.substring(0, result.indexOf(END_TAG));
    }

}
